package edu.washington.cs.knowitall.extractor;

import java.util.Iterator;

import edu.washington.cs.knowitall.extractor.mapper.Mapper;
import edu.washington.cs.knowitall.extractor.mapper.MapperList;
import edu.washington.cs.knowitall.util.AbstractNestedIterator;
import edu.washington.cs.knowitall.util.IterableAdapter;

/**
 * An abstract class that defines the basic behavior of an extractor. An <code>Extractor<S,T></code> object
 * extracts objects of type <code>T</code> from a source object of type <code>S</code>. Candidate extractions
 * are first obtained by calling the <code>extractCandidates(S source)</code> method, which returns an 
 * <code>Iterable</code> object over extractions of type <code>T</code>. These extractions are passed through
 * a list of <code>Mapper<T></code> objects, each of which can filter or modify the extractions. 
 * 
 * Other objects can use an <code>Extractor<S,T></code> object by calling the <code>extract(S source)</code>
 * object, which returns an <code>Iterable<T></code> object of extractions after the <code>Mapper</code>s have
 * been applied. 
 * 
 * <code>Mapper</code> objects can be added to the list of <code>Mapper</code>s by calling the 
 * <code>addMapper(Mapper<T> mapper)</code> method. This will add a <code>Mapper</code> to the end of the list
 * (i.e. it is the last one to be applied to the extractions).
 * 
 * Subclasses extending <code>Extractor<S,T></code> must implement the abstract <code>extractCandidates</code>
 * method. 
 * 
 * As an example, this class can be used to implement a class for extracting <code>String</code> sentences from
 * a <code>String</code> block of text. <code>Mapper</code> objects can be added to filter the sentences by
 * length, or remove brackets from the sentences.
 * 
 * @author afader
 *
 * @param <S>
 * @param <T>
 */
public abstract class Extractor<S, T> {

    private MapperList<T> mappers;

    /**
     * Constructs a new extractor with no <code>Mapper</code> objects.
     */
    public Extractor() {
        mappers = new MapperList<T>();
    }

    /**
     * @return A <code>MapperList<T></code> object containing the mappers assigned to this extractor.
     */
    public MapperList<T> getMappers() {
        return mappers;
    }

    /**
     * Adds a <code>Mapper</code> object to the end of the list of mappers. It will be the new final
     * <code>Mapper</code> object applied to the extractions, after the existing <code>Mapper</code>s
     * have been applied.
     * @param mapper The mapper to add.
     */
    public void addMapper(Mapper<T> mapper) {
        mappers.addMapper(mapper);
    }

    /**
     * Extracts candidate extractions from the given source object. When the user calls the 
     * <code>extract(S source)</code> method, the <code>extractCandidate(S source)</code> method is
     * used to generate a set of candidate extractions, which are then passed through each <code>Mapper</code>
     * object assigned to the extractor. 
     * 
     * @param source The source to extract from.
     * @return An <code>Iterable<T></code> object over the candidate extractions.
     */
    protected abstract Iterable<T> extractCandidates(S source);

    /**
     * @param source the source object to extract from.
     * @return an <code>Iterable<T></code> object over extractions from <code>source</code>.
     */
    public Iterable<T> extract(S source) {
        Iterable<T> candidates = extractCandidates(source);
        return mappers.map(candidates);
    }

    /**
     * @param sources a collection of source objects to extract from.
     * @return an <code>Iterable<T></code> object over extractions from each of the sources.
     */
    public Iterable<T> extract(Iterable<S> sources) {
        Iterator<T> tIter = new AbstractNestedIterator<S,T>(sources) {
            protected Iterator<T> computeInnerIterator(S object) {
                return extract(object).iterator();
            }
        };
        return new IterableAdapter<T>(tIter);
    }

    /**
     * Composes a <code>R->S</code> extractor with a <code>S->T</code> extractor to create
     * a <code>R->T</code> extractor. 
     * @param <R>
     * @param <S>
     * @param <T>
     * @param rsExtractor
     * @param stExtractor
     * @return an extractor taking objects of type <code>R</code> and returning objects of type <code>T</code>
     */
    public static <R,S,T> Extractor<R,T> compose(Extractor<R,S> rsExtractor, Extractor<S,T> stExtractor) {
        return new ExtractorComposition<R,S,T>(rsExtractor, stExtractor);
    }
}
