package edu.washington.cs.knowitall.extractor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import edu.washington.cs.knowitall.nlp.ChunkedSentence;
import edu.washington.cs.knowitall.nlp.PosTaggedPattern;
import edu.washington.cs.knowitall.nlp.extraction.ChunkedExtraction;
import edu.washington.cs.knowitall.util.Range;

/**
 * An extractor that uses a regular expression pattern over POS tags and tokens to extract 
 * relations from NP-chunked sentences. This class uses a <code>PosTaggedPattern</code> object
 * to represent the regular expression pattern.
 * @author afader
 *
 */
public class RegexRelationExtractor extends Extractor<ChunkedSentence, ChunkedExtraction> {

    private String patternString;
    private PosTaggedPattern pattern;

    /**
     * Constructs a new <code>RegexRelationExtractor</code> using the given pattern.
     * @param patternString the relation pattern
     */
    public RegexRelationExtractor(String patternString) {
        this.patternString = patternString;
        this.pattern = new PosTaggedPattern(patternString);
    }

    /**
     * @return the <code>String</code> relation pattern.
     */
    public String getPatternString() {
        return patternString;
    }

    /**
     * @return the <code>PosTaggedPattern</code> used to extract relations.
     */
    public PosTaggedPattern getPattern() {
        return pattern;
    }

    @Override
    /**
     * Extracts relations matching the regular expression returned by <code>getPattern()</code>.
     */
    public Iterable<ChunkedExtraction> extractCandidates(ChunkedSentence sentence) {

        List<Range> ranges = pattern.findAll(sentence);

        Collection<ChunkedExtraction> results = new ArrayList<ChunkedExtraction>(); 
        for (Range range : ranges) {
            ChunkedExtraction extr = new ChunkedExtraction(sentence, range);
            results.add(extr);
        }

        return results;
    }

}
