package edu.washington.cs.knowitall.nlp;

import java.util.Arrays;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import com.google.common.collect.ImmutableList;

import edu.washington.cs.knowitall.util.Pair;
import edu.washington.cs.knowitall.util.Range;

/**
 * An immutable class that represents a tokenized, POS-tagged, and noun-phrase chunked sentence. 
 * @author afader
 *
 */
public class ChunkedSentence implements PosTagged, ChunkedSource {

	private final ImmutableList<String> tokens;
    private final ImmutableList<String> posTags;
    private final ImmutableList<String> npChunkTags;

    private final ImmutableList<Range> npChunkRanges;
    
    // a cache for getTokensAsString
    private String tokensAsString = null;

    /**
     * Constructs a new <code>NpChunkedSentence</code> object using the given tokens, POS tags, and
     * NP chunk tags, each of which must have the same length. The NP chunks should be expressed using
     * the standard <code>B-NP</code>, <code>I-NP</code>, <code>O</code> tags.
     * @param tokens
     * @param posTags
     * @param npChunkTags
     */
    public ChunkedSentence(String[] tokens, String[] posTags, String[] npChunkTags) {
        this(Arrays.asList(tokens), Arrays.asList(posTags), Arrays.asList(npChunkTags));
    }
    
    public ChunkedSentence(List<String> tokens, List<String> posTags, List<String> npChunkTags) {
        if (!(tokens.size() == posTags.size() && posTags.size() == npChunkTags.size())) {
            throw new IllegalArgumentException("Tokens, POS tags, and NP chunk tags must be the same length. Got " + tokens.size() + ", " + posTags.size() + ", " + npChunkTags.size());
        } else {
            this.tokens = ImmutableList.copyOf(tokens);
            this.posTags = ImmutableList.copyOf(posTags);
            this.npChunkTags = ImmutableList.copyOf(npChunkTags);
            this.npChunkRanges = ImmutableList.copyOf(OpenNlpUtils.computeNpChunkRanges(npChunkTags));
        }
    }

    /**
     * Constructs a new <code>NpChunkedSentence</code> by copying the data of <code>sent</code>.
     * @param sent
     */
    public ChunkedSentence(ChunkedSentence sent) {
        this.tokens = sent.tokens;
        this.posTags = sent.posTags;
        this.npChunkTags = sent.npChunkTags;
        this.npChunkRanges = sent.npChunkRanges;
    }

    /**
     * Returns a copy of this <code>NpChunkedSentence</code>.
     */
    public ChunkedSentence clone() {
        return new ChunkedSentence(tokens, posTags, npChunkTags);
    }

    /**
     * @return the number of noun phrase chunks in this sentence.
     */
    public int getNumNpChunks() {
        return npChunkRanges.size();
    }

    /**
     * @return the number of tokens in this sentence.
     */
    public int getLength() {
        return this.tokens.size();
    }

    /**
     * @return an unmodifiable list over the tokens of this sentence.
     */
    public List<String> getTokens() {
        return this.tokens;
    }

    /**
     * @return an unmodifiable list over the POS tags of this sentence.
     */
    public List<String> getPosTags() {
        return this.posTags;
    }

    /**
     * @param start
     * @param length
     * @return the first <code>length</code> tokens starting at index <code>start</code>. 
     */
    public List<String> getTokens(int start, int length) {
        return this.tokens.subList(start, start + length);
    }

    /**
     * @param range
     * @return the tokens at the indexes given by <code>range</code>.
     */
    public List<String> getTokens(Range range) {
        return getTokens(range.getStart(), range.getLength());
    }

    /**
     * @param start
     * @param length
     * @return the first <code>length</code> POS tags starting at index <code>start</code>. 
     */
    public List<String> getPosTags(int start, int length) {
        return this.posTags.subList(start, start + length);
    }

    /**
     * @param range
     * @return the POS tags at the indexes given by <code>range</code>.
     */
    public List<String> getPosTags(Range range) {
        return getPosTags(range.getStart(), range.getLength());
    }

    /**
     * @return an unmodifiable list over the ranges of the NP chunks in this sentence.
     */
    public Iterable<Range> getNpChunkRanges() {
        return npChunkRanges;
    }

    /**
     * @return an unmodifiable list over the NP chunk tags of this sentence.
     */
    public List<String> getNpChunkTags() {
        return this.npChunkTags;
    }
    
    /**
     * @param start
     * @param length
     * @return the first <code>length</code> NP chunk tags starting at index <code>start</code>. 
     */
    public List<String> getNpChunkTags(int start, int length) {
    	return this.npChunkTags.subList(start, start + length);
    }
    
    /**
     * @param range
     * @return the first <code>length</code> NP chunk tags in the range <code>range</code>.
     */
    public List<String> getNpChunkTags(Range range) {
    	return getNpChunkTags(range.getStart(), range.getLength());
    }
    
    /**
     * @param i
     * @return the NP chunk tag at index <code>i</code>.
     */
    public String getNpChunkTag(int i) {
    	return this.npChunkTags.get(i);
    }

    /**
     * @return a reference to this object.
     */
    public ChunkedSentence getSentence() {
        return this;
    }

    /**
     * @return the tokens of this sentence joined by spaces.
     */
    public String getTokensAsString() {
        return getTokensAsString(0, getLength());
    }

    /**
     * @param start
     * @param length
     * @return <code>length</code> tokens starting at <code>start</code>, joined by spaces.
     */
    public String getTokensAsString(int start, int length) {
        return StringUtils.join(getTokens(start, length).iterator(), " ");
    }

    /**
     * @param range
     * @return the tokens at the indexes of <code>range</code>, joined by spaces.
     */
    public String getTokensAsString(Range range) {
        if (this.tokensAsString == null) {
            this.tokensAsString = getTokensAsString(range.getStart(), range.getLength());
        }
        
        return this.tokensAsString;
    }

    /**
     * @param start
     * @param length
     * @return <code>length</code> POS tags starting at <code>start</code>, joined by spaces.
     */
    public String getPosTagsAsString(int start, int length) {
        return StringUtils.join(getPosTags(start, length).iterator(), " ");
    }

    /**
     * @param range
     * @return the POS tags at the indexes of <code>range</code>, joined by spaces.
     */
    public String getPosTagsAsString(Range range) {
        return getPosTagsAsString(range.getStart(), range.getLength());
    }

    /**
     * @return the POS tags of this sentence, joined by spaces.
     */
    public String getPosTagsAsString() {
        return getPosTagsAsString(0, getLength());
    }

    /**
     * @return the NP chunk tags of this sentence (in B-NP, I-NP, O format), joined by strings.
     */
    public String getNpChunkTagsAsString() {
        String[] result = new String[getLength()];
        for (int i = 0; i < getLength(); i++) result[i] = "O";
        for (Range r : getNpChunkRanges()) {
            for (int i : r) result[i] = "I-NP";
            result[r.getStart()] = "B-NP";
        }
        return StringUtils.join(result, " ");
    }

    /**
     * @return a string representation of this sentence.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < getLength(); i++) {

            String tt = this.tokens.get(i) + "/" + this.posTags.get(i);
            String npChunkTag = this.npChunkTags.get(i);
            if (i > 0 && !npChunkTag.startsWith("I-") && !this.npChunkTags.get(i-1).equals("O")) {
                sb.append("]");
            }
            if (npChunkTag.startsWith("B-")) {
                sb.append(" [" + npChunkTag.substring(2));
            }
            sb.append(" " + tt);
        }
        return sb.toString();
    }

    @Override
    /**
     * @return a pair object containing the (token, POS tag) pair at index i.
     */
    public Pair<String, String> getPair(int i) {
        return new Pair<String, String>(this.tokens.get(i), this.posTags.get(i));
    }

	@Override
	public String getToken(int i) {
		return this.tokens.get(i);
	}

	@Override
	public String getPosTag(int i) {
		return this.posTags.get(i);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((npChunkRanges == null) ? 0 : npChunkRanges.hashCode());
		result = prime * result
				+ ((npChunkTags == null) ? 0 : npChunkTags.hashCode());
		result = prime * result + ((posTags == null) ? 0 : posTags.hashCode());
		result = prime * result + ((tokens == null) ? 0 : tokens.hashCode());
		result = prime * result
				+ ((tokensAsString == null) ? 0 : tokensAsString.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ChunkedSentence other = (ChunkedSentence) obj;
		if (npChunkRanges == null) {
			if (other.npChunkRanges != null)
				return false;
		} else if (!npChunkRanges.equals(other.npChunkRanges))
			return false;
		if (npChunkTags == null) {
			if (other.npChunkTags != null)
				return false;
		} else if (!npChunkTags.equals(other.npChunkTags))
			return false;
		if (posTags == null) {
			if (other.posTags != null)
				return false;
		} else if (!posTags.equals(other.posTags))
			return false;
		if (tokens == null) {
			if (other.tokens != null)
				return false;
		} else if (!tokens.equals(other.tokens))
			return false;
		if (tokensAsString == null) {
			if (other.tokensAsString != null)
				return false;
		} else if (!tokensAsString.equals(other.tokensAsString))
			return false;
		return true;
	}

}
