package edu.washington.cs.knowitall.nlp;

import java.io.IOException;
import java.util.Date;
import java.util.regex.Pattern;

import edu.washington.cs.knowitall.util.DefaultObjects;

import opennlp.tools.chunker.Chunker;
import opennlp.tools.postag.POSTagger;
import opennlp.tools.tokenize.Tokenizer;

/**
 * A class that combines OpenNLP tokenizer, POS tagger, and chunker objects into a single
 * object that converts <code>String</code> sentences to <code>NpChunkedSentence</code>
 * objects. By default, uses the models from <code>DefaultObjects.getDefaultTokenizer()</code>,
 * <code>DefaultObjects.getDefaultPosTagger()</code>, and <code>DefaultObjects.getDefaultChunker()</code>.
 * @author afader
 *
 */
public class OpenNlpSentenceChunker implements SentenceChunker {

    private Chunker chunker;
    private POSTagger posTagger;
    private Tokenizer tokenizer;
    
    private OpenNlpSentenceChunkerStats lastChunkedSentenceStats = null;

    private boolean attachOfs = true;
    private boolean attachPossessives = true;
    
    Pattern convertToSpace = Pattern.compile("\\xa0");

    /**
     * Constructs a new <code>OpenNlpSentenceChunker</code> object using the default models from
     * <code>DefaultObjects</code>.
     * @throws IOException if unable to load the models.
     */
    public OpenNlpSentenceChunker() throws IOException {
        this.tokenizer = DefaultObjects.getDefaultTokenizer();
        this.posTagger = DefaultObjects.getDefaultPosTagger();
        this.chunker = DefaultObjects.getDefaultChunker();
    }

    /**
     * Constructs a new <code>OpenNlpSentenceChunker</code> object using the provided OpenNLP objects.
     * @param tokenizer
     * @param posTagger
     * @param chunker
     */
    public OpenNlpSentenceChunker(Tokenizer tokenizer, POSTagger posTagger, Chunker chunker) {
        this.tokenizer = tokenizer;
        this.posTagger = posTagger;
        this.chunker = chunker;
    }

    /**
     * @return true if this object will attach NPs beginning with "of" with the previous NP.
     */
    public boolean attachOfs() {
        return attachOfs;
    }

    /**
     * @return true if this object will attach NPs beginning with the tag POS with the previous NP.
     */
    public boolean attachPossessives() {
        return attachPossessives;
    }

    /**
     * @param attachOfs
     */
    public void attachOfs(boolean attachOfs) {
        this.attachOfs = attachOfs;
    }

    /**
     * @param attachPossessives
     */
    public void attachPossessives(boolean attachPossessives) {
        this.attachPossessives = attachPossessives;
    }
    
    /**
     * Returns statistics about the last call to chunkSentence or null.
     */
    public OpenNlpSentenceChunkerStats lastChunkedSentenceStats() {
        return this.lastChunkedSentenceStats;
    }

    @Override
    /**
     * Chunks the given sentence and returns it as an <code>NpChunkedSentence</code> object.
     */
    public ChunkedSentence chunkSentence(String sent) {
        
        // OpenNLP cannot handle non-breaking whitespace
        sent = convertToSpace.matcher(sent).replaceAll(" ");

        Date overallStart = new Date();
        Date start;
        
        start = new Date();
        String[] tokens = tokenizer.tokenize(sent);
        long tokenizeTime = (new Date().getTime()) - start.getTime();
        
        start = new Date();
        String[] posTags = posTagger.tag(tokens);
        long tagTime = (new Date().getTime()) - start.getTime();
        
        start = new Date();
        String[] npChunkTags = chunker.chunk(tokens, posTags);
        long chunkTime = (new Date().getTime()) - start.getTime();

        if (attachOfs) OpenNlpUtils.attachOfs(tokens, npChunkTags);
        if (attachPossessives) OpenNlpUtils.attachPossessives(posTags, npChunkTags);
        
        long totalTime = (new Date().getTime()) - overallStart.getTime();
        this.lastChunkedSentenceStats = new OpenNlpSentenceChunkerStats(totalTime, tokenizeTime, tagTime, chunkTime);
        
        ChunkedSentence result = new ChunkedSentence(tokens, posTags, npChunkTags);
        return result;
    }

}
