package edu.washington.cs.knowitall.nlp;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.washington.cs.knowitall.util.Pair;
import edu.washington.cs.knowitall.util.Range;

public class PosTaggedPattern {

    private static final Pattern TOKEN_PATTERN = Pattern.compile("([^\\\\\\[\\](){}^.?*+|\\s]+)");
    private static final HashSet<Character> METACHARS = new HashSet<Character>(Arrays.asList(new Character[] {
            '\\', '^', '$', '?', '*', '+', '[', ']', '{', '}', '(', ')', '.', '|', '-', '&', ':'}));
    private static final String POS_TAG_PREFIX = "POS_";
    private static final String UNK_TOKEN = "UNK_TOKEN";
    private static final String UNK_POS_TAG = "UNK_POS_TAG";
    private static final int START_CHAR = 0x0041;
    private static final int MAX_CHAR = 0xFFFF-1;

    private HashSet<String> posTagSet;
    private HashSet<String> tokenSet;
    private HashMap<Pair<String, String>, Character> pairEncodingTable;
    private Pattern encodedPattern;

    private static HashSet<String> tokenizeInputPattern(String inputPattern) {
        Matcher m = TOKEN_PATTERN.matcher(inputPattern);
        HashSet<String> results = new HashSet<String>();
        while (m.find()) {
            results.add(m.group());
        }
        return results;
    }

    private void createTagSets(HashSet<String> set) {
        posTagSet = new HashSet<String>();
        tokenSet = new HashSet<String>();
        for (String s : set) {
            if (s.startsWith(POS_TAG_PREFIX)) {
                posTagSet.add(s.substring(POS_TAG_PREFIX.length()));
            } else {
                tokenSet.add(s);
            }
        }
        posTagSet.add(UNK_POS_TAG);
        tokenSet.add(UNK_TOKEN);
    }

    private static char unicode(int i) {
        if (i > MAX_CHAR) {
            throw new IllegalArgumentException("Cannot encode " + i);
        } else {
            char[] things = Character.toChars(i);
            assert things.length == 1;
            return things[0];
        }
    }

    private void createPairEncodingTable() {
        int numTokens = tokenSet.size();
        int numPosTags = posTagSet.size();
        int tableSize = numTokens * numPosTags;
        pairEncodingTable = new HashMap<Pair<String,String>, Character>(tableSize);
        int i = START_CHAR;
        for (String token : tokenSet) {
            for (String posTag : posTagSet) {
                Pair<String,String> pair = new Pair<String, String>(token, posTag);
                while (METACHARS.contains(unicode(i))) i++;
                pairEncodingTable.put(pair, unicode(i));
                i++;
            }
        }
    }

    private char getEncoding(String token, String posTag) {
        if (!tokenSet.contains(token)) {
            token = UNK_TOKEN;
        }
        if (!posTagSet.contains(posTag)) {
            posTag = UNK_POS_TAG;
        }
        Pair<String,String> pair = new Pair<String,String>(token, posTag);
        if (pairEncodingTable.containsKey(pair)) {
            return pairEncodingTable.get(pair);
        } else {
            throw new IllegalArgumentException("Cannot file pair " + pair + " in table.");
        }
    }

    private String anyTokenCharClass(String posTag) {
        StringBuffer chars = new StringBuffer();
        for (String token : tokenSet) {
            chars.append(getEncoding(token, posTag));
        }
        return "["+chars.toString()+"]";
    }

    private String anyPosTagCharClass(String token) {
        StringBuffer chars = new StringBuffer();
        for (String posTag : posTagSet) {
            chars.append(getEncoding(token, posTag));
        }
        return "["+chars.toString()+"]";
    }

    private String substituteInSection(String section) {
        for (String token : tokenSet) {
            String s = " " + Pattern.quote(token) + " ";
            section = section.replaceAll(s, " " + anyPosTagCharClass(token) + " ");
        }

        for (String posTag : posTagSet) {
            String s = " " + Pattern.quote(POS_TAG_PREFIX + posTag) + " ";
            section = section.replaceAll(s, " " + anyTokenCharClass(posTag) + " ");
        }     
        return section;
    }

    protected static String fixBrackets(String s) {
        s = s.replaceAll("\\[\\^\\[", "[^");
        s = s.replaceAll("\\[\\[", "[");
        s = s.replaceAll("\\]\\]", "]");
        s = s.replaceAll("\\]\\[", "");
        return s;
    }

    private void encodePattern(String pattern) {
        pattern = TOKEN_PATTERN.matcher(pattern).replaceAll(" $1 ");
        HashSet<String> tokens = tokenizeInputPattern(pattern);
        createTagSets(tokens);
        createPairEncodingTable();

        Pattern bracketPattern = Pattern.compile("(\\[\\^?[^\\]]+\\]|[^\\[\\]]+)");
        Matcher m = bracketPattern.matcher(pattern);
        StringBuilder result = new StringBuilder();
        while (m.find()) {
            String section = m.group();
            String substitute = substituteInSection(section);
            substitute = substitute.replaceAll("\\s+", "");
            if (section.startsWith("[")) {
                substitute = fixBrackets(substitute);
            }
            result.append(substitute);
        }

        pattern = result.toString();
        pattern = pattern.replaceAll("\\s+", "");
        this.encodedPattern = Pattern.compile(pattern);
    }

    private String encodePosTagged(PosTagged sent) {
        int n = sent.getLength();
        char[] values = new char[n];
        for (int i = 0; i < n; i++) {
            String token = sent.getTokens().get(i);
            String posTag = sent.getPosTags().get(i);
            values[i] = getEncoding(token, posTag);
        }
        return new String(values);
    }

    public List<Range> findAll(PosTagged sent) {
        String encodedSent = encodePosTagged(sent);
        Matcher m = encodedPattern.matcher(encodedSent);
        ArrayList<Range> ranges = new ArrayList<Range>();
        while (m.find()) {
            int start = m.start();
            int end = m.end();
            ranges.add(new Range(start, end-start));
        }
        return ranges;
    }

    public String toString() {
        return encodedPattern.toString();
    }

    private void init(String inputPattern) {
        encodePattern(inputPattern);
    }

    public PosTaggedPattern(String inputPattern) {
        init(inputPattern);
    }


}

