package edu.washington.cs.knowitall.nlp.extraction;

import java.util.ArrayList;
import java.util.Collection;

import org.apache.commons.lang.StringUtils;

import edu.washington.cs.knowitall.nlp.ChunkedSentence;

/**
 * An <code>Extraction</code> representing a binary extraction with noun phrase arguments. The relation is
 * represented as a <code>NpChunkedExtraction</code> object and the arguments are represented as 
 * <code>NpChunkArgumentExtraction</code> objects.
 * @author afader
 *
 */
public class ChunkedBinaryExtraction implements Extraction<ChunkedSentence>{

    private ChunkedExtraction relation;
    private ChunkedArgumentExtraction argument1;
    private ChunkedArgumentExtraction argument2;

    /**
     * Constructs a collection of <code>NpChunkedBinaryExtraction</code> objects with arguments given by
     * the cross product of <code>arg1s</code> with <code>arg2s</code>, and relation <code>rel</code>.
     * @param rel
     * @param arg1s
     * @param arg2s
     * @return a collection of binary extractions with relation <code>rel</code> and arguments from <code>arg1s</code> and <code>arg2s</code>.
     */
    public static Collection<ChunkedBinaryExtraction> productOfArgs(ChunkedExtraction rel, Iterable<? extends ChunkedArgumentExtraction> arg1s, Iterable<? extends ChunkedArgumentExtraction> arg2s) {
        Collection<ChunkedBinaryExtraction> results = new ArrayList<ChunkedBinaryExtraction>();
        for (ChunkedArgumentExtraction arg1 : arg1s) {
            for (ChunkedArgumentExtraction arg2 : arg2s) {
                ChunkedBinaryExtraction extr = new ChunkedBinaryExtraction(rel, arg1, arg2);
                results.add(extr);
            }
        }
        return results;
    }

    @Override
    /**
     * @return the source of this extraction.
     */
    public ChunkedSentence getSource() {
        return relation.getSource();
    }

    /**
     * Constructs a new <code>NpChunkedBinaryExtraction</code> with the given relation and arguments.
     * @param relation
     * @param argument1
     * @param argument2
     */
    public ChunkedBinaryExtraction(ChunkedExtraction relation, ChunkedArgumentExtraction argument1, ChunkedArgumentExtraction argument2) {
        this.relation = relation;
        this.argument1 = argument1;
        this.argument2 = argument2;
    }

    /**
     * @return the first argument of this extraction.
     */
    public ChunkedArgumentExtraction getArgument1() {
        return argument1;
    }

    /**
     * @return the relation of this extraction.
     */
    public ChunkedExtraction getRelation() {
        return relation;
    }

    /**
     * @return the second argument of this extraction.
     */
    public ChunkedArgumentExtraction getArgument2() {
        return argument2;
    }
    
    /**
     * @return the sentence this extraction came from 
     */
    public ChunkedSentence getSentence() {
    	return relation.getSentence();
    }

    @Override
    /**
     * @return a <code>String</code> in the form (arg1, rel, arg2).
     */
    public String toString() {
        String[] fields = new String[] { getArgument1().toString(), getRelation().toString(), getArgument2().toString() };
        return "(" + StringUtils.join(fields, ", ") + ")";
    }

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((argument1 == null) ? 0 : argument1.hashCode());
		result = prime * result
				+ ((argument2 == null) ? 0 : argument2.hashCode());
		result = prime * result
				+ ((relation == null) ? 0 : relation.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ChunkedBinaryExtraction other = (ChunkedBinaryExtraction) obj;
		if (argument1 == null) {
			if (other.argument1 != null)
				return false;
		} else if (!argument1.equals(other.argument1))
			return false;
		if (argument2 == null) {
			if (other.argument2 != null)
				return false;
		} else if (!argument2.equals(other.argument2))
			return false;
		if (relation == null) {
			if (other.relation != null)
				return false;
		} else if (!relation.equals(other.relation))
			return false;
		return true;
	}

}
