package edu.washington.cs.knowitall.nlp.extraction;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import edu.washington.cs.knowitall.nlp.ChunkedSentence;
import edu.washington.cs.knowitall.nlp.ChunkedSource;
import edu.washington.cs.knowitall.nlp.PosTagged;
import edu.washington.cs.knowitall.util.Pair;
import edu.washington.cs.knowitall.util.Range;

/**
 * An <code>Extraction</code> object that represents a contiguous subsequence of a <code>NpChunkedSentence</code>
 * object. A <code>NpChunkedExtraction</code> has a reference to the <code>NpChunkedSentence</code> object <code>sent</code>
 * and a <code>Range</code> object <code>range</code>. <code>range</code> is the subsequence of <code>sent</code> that
 * this extraction represents.
 * @author afader
 *
 */
public class ChunkedExtraction 
    implements PosTagged, Extraction<ChunkedSentence>, ChunkedSource {

    private Range range;
    private ChunkedSentence sent;
    private List<Range> npChunkRanges;
    
    // a cache for getTokensAsString
    private String tokensAsString = null;

    /**
     * Constructs a new <code>NpChunkedExtraction</code> object representing <code>range</code> in 
     * the sentence <code>sent</code>. <code>range</code> must be a subset of [0, <code>sent.getLength()</code>).
     * @param sent the source sentence.
     * @param range the subsequence of <code>sent</code> that this extraction will represent.
     */
    public ChunkedExtraction(ChunkedSentence sent, Range range) {

        int start = range.getStart();
        int length = range.getLength();
        int sentLength = sent.getLength();

        if (start >= sentLength) {
            throw new IllegalArgumentException("Invalid start position for chunked extraction on sentence: pos=" + start);
        }

        if (start + length > sentLength) {
            throw new IllegalArgumentException("Invalid length for sentence: " + range.getLength());
        }

        this.range = range;
        this.sent = sent;
        this.npChunkRanges = new ArrayList<Range>();
        for (Range npChunkRange : sent.getNpChunkRanges()) {
            int chunkStart = npChunkRange.getStart();
            int chunkLength = npChunkRange.getLength();
            if (start <= chunkStart && chunkStart < (start + length) && (chunkStart + chunkLength) <= (start + length)) {
                npChunkRanges.add(npChunkRange);
            }
        }

    }

    /**
     * @return the start index of the extraction.
     */
    public int getStart() {
        return getRange().getStart();
    }

    /**
     * @return the <code>Range</code> object that represents the 
     * subsequence of the source sentence.
     */
    public Range getRange() {
        return range;
    }

    @Override
    /**
     * @return the source sentence of this extraction. 
     */
    public ChunkedSentence getSource() {
        return sent;
    }

    /**
     * @return the source sentence of this extraction.
     */
    public ChunkedSentence getSentence() {
        return sent;
    }

    @Override
    /**
     * @return the length of this extraction, in number of words.
     */
    public int getLength() {
        return range.getLength();
    }

    /**
     * @param start
     * @param length
     * @return true if [start, start+length) is a subsequence of this 
     * extraction's range.
     */
    private boolean isValidRange(int start, int length) {
        boolean startOk = 0 <= start && start < getLength();
        boolean lengthOk = 0 < length && length <= getLength();
        return startOk && lengthOk;
    }

    /**
     * @return the tokens of this extraction.
     */
    public List<String> getTokens() {
        return sent.getTokens(getRange());
    }

    @Override
    /**
     * @return the tokens of this extraction, starting at start and 
     * ending at start+length-1.
     */
    public List<String> getTokens(int start, int length) {
        if (isValidRange(start, length)) { 
            start = getStart() + start;
            return sent.getTokens(start, length);
        } else {
            throw new IllegalArgumentException("Range out of bounds: " + 
                    start + ", " + length);
        }
    }
    
    /**
     * @return the tokens joined by space.
     */
    public String getTokensAsString() {
        if (this.tokensAsString == null) {
            this.tokensAsString = StringUtils.join(getTokens().iterator(), ' ');
        }
        
        return this.tokensAsString;
    }

    @Override
    /**
     * @param range
     * @return the tokens of this extraction in the given range.
     */
    public List<String> getTokens(Range range) {
        return getTokens(getStart(), getLength());
    }

    @Override
    /**
     * @return the POS tags of this extraction.
     */
    public List<String> getPosTags() {
        return sent.getPosTags(getRange());
    }

    @Override
    /**
     * @param start
     * @param length
     * @return the POS tags at positions start to start+length-1 of this 
     * extraction.
     */
    public List<String> getPosTags(int start, int length) {
        if (isValidRange(start, length)) {
            return sent.getPosTags(start + getStart(), length);
        } else {
            throw new IllegalArgumentException("Range out of bounds: " + 
                    start + ", " + length);
        }
    }

    /**
     * @param extr
     * @return true if this extraction is adjacent to or overlaps with 
     * extr in this sentence.
     */
    public boolean isAdjacentOrOverlaps(ChunkedExtraction extr) {
        return getRange().isAdjacentOrOverlaps(extr.getRange());
    }

    /**
     * @param extr
     * @return true if this extraction overlaps with extr in this sentence.
     */
    public boolean overlapsWith(ChunkedExtraction extr) {
        return getRange().overlapsWith(extr.getRange());
    }

    @Override
    /**
     * @param range
     * @return the POS tags of this extraction at the positions given by 
     * range.
     */
    public List<String> getPosTags(Range range) {
        return getPosTags(getStart(), getLength());
    }

    /**
     * @return the tokens of this extraction, joined by a space.
     */
    public String toString() {
        return StringUtils.join(getTokens().iterator(), ' ');
    }

    @Override
    /**
     * @param i
     * @return a pair of strings representing the (token, POS tag) at 
     * position i of this extraction.
     */
    public Pair<String, String> getPair(int i) {
        String token = this.getTokens().get(i);
        String posTag = this.getPosTags().get(i);
        return new Pair<String, String>(token, posTag);
    }

	@Override
	public String getToken(int i) {
		return this.getTokens().get(i);
	}

	@Override
	public String getPosTag(int i) {
		return this.getPosTags().get(i);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((npChunkRanges == null) ? 0 : npChunkRanges.hashCode());
		result = prime * result + ((range == null) ? 0 : range.hashCode());
		result = prime * result + ((sent == null) ? 0 : sent.hashCode());
		result = prime * result
				+ ((tokensAsString == null) ? 0 : tokensAsString.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ChunkedExtraction other = (ChunkedExtraction) obj;
		if (npChunkRanges == null) {
			if (other.npChunkRanges != null)
				return false;
		} else if (!npChunkRanges.equals(other.npChunkRanges))
			return false;
		if (range == null) {
			if (other.range != null)
				return false;
		} else if (!range.equals(other.range))
			return false;
		if (sent == null) {
			if (other.sent != null)
				return false;
		} else if (!sent.equals(other.sent))
			return false;
		if (tokensAsString == null) {
			if (other.tokensAsString != null)
				return false;
		} else if (!tokensAsString.equals(other.tokensAsString))
			return false;
		return true;
	}

}
