package test.edu.washington.cs.knowitall.nlp;

import java.util.ArrayList;
import java.util.List;

import junit.framework.Assert;

import org.junit.Before;
import org.junit.Test;

import edu.washington.cs.knowitall.nlp.ChunkedSentence;
import edu.washington.cs.knowitall.nlp.PosTaggedPattern;
import edu.washington.cs.knowitall.util.Range;

public class PosTaggedPatternTest {

    private ChunkedSentence sent1, sent2, sent3, sent4, sent5;
    List<Range> expectedRanges, gotRanges;

    @Before
    public void setUp() throws Exception {

        sent1 = new ChunkedSentence(
                new String[] { "I", "saw", "the", "dog", "." },
                new String[] { "PP", "VBD", "DT", "NN", "." },
                new String[] { "B-NP", "O", "B-NP", "I-NP", "." }
        );

        sent2 = new ChunkedSentence(
                new String[] { "He", "went", "to", "the", "store", "to", "buy", "some", "eggs", "!" },
                new String[] { "PP", "VBD", "IN", "DT", "NN", "TO", "VB", "DT", "NNS", "." },
                new String[] { "B-NP", "O", "O", "B-NP", "I-NP", "O", "O", "B-NP", "I-NP", "O" }
        );

        sent3 = new ChunkedSentence(
                new String[] { "Obama", "was", "a", "professor", "of", "Law", "at", "UChicago", "." },
                new String[] { "NNP", "VBD", "DT", "NN", "IN", "NN", "IN", "NNP", "." },
                new String[] { "B-NP", "O", "B-NP", "I-NP", "I-NP", "I-NP", "O", "B-NP", "O" }
        );

        sent4 = new ChunkedSentence(
                new String[] { "Joe", "lost", "his", "job", "at", "the", "plant", "."},
                new String[] { "NNP", "VBD", "PRP$", "NN", "IN", "DT", "NN", "."},
                new String[] { "B-NP", "O", "B-NP", "I-NP", "O", "B-NP", "I-NP", "O"}
        );

        sent5 = new ChunkedSentence(
                "I like this , but not that .".split(" "),
                "PRP VB DT , CC RB DT .".split(" "),
                "B-NP O B-NP O O O B-NP O".split(" ")
        );

        expectedRanges = new ArrayList<Range>();
        gotRanges = new ArrayList<Range>();
    }

    @Test
    public void testFindAll1() {
        PosTaggedPattern pattern = new PosTaggedPattern("POS_DT POS_NN");
        expectedRanges.add(new Range(2,2));
        List<Range> gotRanges = pattern.findAll(sent1);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll2() {
        PosTaggedPattern pattern = new PosTaggedPattern("POS_VBD POS_DT dog");
        expectedRanges.add(new Range(1,3));
        gotRanges = pattern.findAll(sent1);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll3() {
        PosTaggedPattern pattern = new PosTaggedPattern("POS_DT (POS_NN|POS_NNS)");
        expectedRanges.add(new Range(3, 2));
        expectedRanges.add(new Range(7, 2));
        gotRanges = pattern.findAll(sent2);
        Assert.assertEquals(expectedRanges, gotRanges);
    }     

    @Test
    public void testFindAll4() {
        PosTaggedPattern pattern = new PosTaggedPattern("I .* dog");
        expectedRanges.add(new Range(0,4));
        gotRanges = pattern.findAll(sent1);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll5() {
        // Wildcard pattern ".*" will return two ranges, the entire sentence and then 
        // an empty string starting at the end of the sentence.
        PosTaggedPattern pattern = new PosTaggedPattern(".*");
        expectedRanges.add(new Range(0,5));
        expectedRanges.add(new Range(5,0));
        gotRanges = pattern.findAll(sent1);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll6() {
        PosTaggedPattern pattern = new PosTaggedPattern("(POS_VB|POS_VBD) [^the] .");
        gotRanges = pattern.findAll(sent1);
        Assert.assertEquals(expectedRanges, gotRanges);
        expectedRanges.clear();
        expectedRanges.add(new Range(1,3)); // went to the
        expectedRanges.add(new Range(6,3)); // buy some eggs
        gotRanges = pattern.findAll(sent2);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll7() {
        PosTaggedPattern pattern = new PosTaggedPattern("(POS_VB|POS_VBD) [^the to some] .");
        gotRanges = pattern.findAll(sent1);
        Assert.assertEquals(expectedRanges, gotRanges);

        gotRanges = pattern.findAll(sent2);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll8() {
        // Obama was a professor of Law at UChicago.
        PosTaggedPattern pattern = new PosTaggedPattern("POS_VBD [^POS_VBD POS_IN]* POS_IN");
        expectedRanges.add(new Range(1,4));
        gotRanges = pattern.findAll(sent3);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll9() {
        // Joe lost his job at the plant.
        PosTaggedPattern pattern = new PosTaggedPattern("POS_VBD POS_PRP$ POS_NN");
        expectedRanges.add(new Range(1,3));
        gotRanges = pattern.findAll(sent4);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testFindAll10() {
        PosTaggedPattern pattern = new PosTaggedPattern("POS_VBD [^POS_VBD POS_IN]+ of [^POS_VBD POS_IN]+ at");
        expectedRanges.add(new Range(1,6));
        gotRanges = pattern.findAll(sent3);
        Assert.assertEquals(expectedRanges, gotRanges);
    }

    @Test
    public void testNonAlphaNumPattern() {
        PosTaggedPattern pattern = new PosTaggedPattern("like this [^, .]");
        gotRanges = pattern.findAll(sent5);
        Assert.assertEquals(expectedRanges, gotRanges);

    }

    @Test
    public void testWordPattern() {
        String word = "[POS_$ POS_PRP$ POS_CD POS_DT POS_JJ POS_JJS POS_JJR POS_NN POS_NNS POS_NNP POS_NNPS POS_POS POS_PRP POS_RB POS_RBR POS_RBS]";
        PosTaggedPattern pattern = new PosTaggedPattern(word);
        gotRanges = pattern.findAll(sent3);
        expectedRanges.add(new Range(0,1));
        expectedRanges.add(new Range(2,1));
        expectedRanges.add(new Range(3,1));
        expectedRanges.add(new Range(5,1));
        expectedRanges.add(new Range(7,1));
        Assert.assertEquals(expectedRanges, gotRanges);
    }

}
